﻿using System;
using System.Collections.Generic;
using System.Diagnostics;
using System.Globalization;
using System.Reflection;
using JetBrains.Annotations;
using Microsoft.VisualStudio.TestTools.UnitTesting;
using OpcLabs.BaseLib.Execution;
using OpcLabs.BaseLib.Extensions;
using OpcLabs.BaseLib.Extensions.Internal;
using OpcLabs.BaseLib.Reflection.Extensions;
using OpcLabs.BaseLib.Testing;
using OpcLabs.BaseLib.Utilities;

namespace OpcLabs.UnitTestLib.Testers
{
    // IMPROVE: also have InvalidParameterTester

    public class InvalidValueDataMemberTester : Tester
    {
        private class Worker : MarshalByRefObject
        {
            public void Test(
                [NotNull] Type typeUnderTest,
                [NotNull] MemberInfo dataMember,
                [CanBeNull] object value,
                [NotNull] string expectedExceptionTypeFullName)
            {
                object instance = null;
                if (dataMember.IsStatic())
                    AssertUtilities.WriteLine("Data member is static.");
                else
                {
                    AssertUtilities.WriteLine("Creating instance...");
                    instance = Activator.CreateInstance(typeUnderTest, nonPublic: true);
                }
                AssertUtilities.WriteLine("Assigning value to a data member...");
                try
                {
                    dataMember.SetValue(instance, value, null);
                }
                catch (TargetInvocationException targetInvocationException)
                {
                    Exception exception = targetInvocationException.InnerException;
                    Debug.Assert(exception != null);
                    Type actualExceptionType = exception.GetType();
                    string actualExceptionTypeFullName = actualExceptionType.FullName;
                    AssertUtilities.WriteLine(String.Format(CultureInfo.InvariantCulture,
                        "Exception thrown, actual type full name: {0}.",
                        actualExceptionTypeFullName));
                    AssertUtilities.WriteLine(exception);
                    Assert.AreEqual(expectedExceptionTypeFullName, actualExceptionTypeFullName);
                    return;
                }
                Assert.Fail("No exception thrown.");
            }
        }

        public InvalidValueDataMemberTester() 
        {
            TestCategories.Add("DataMembers");
        }

        [NotNull]
        static private TestMethod CreateAssignmentTestMethod(
            [NotNull] Type typeUnderTest,
            [NotNull] MemberInfo dataMember,
            [NotNull] string objectName,
            [NotNull] Func<object> func,
            [NotNull] string expectedExceptionTypeFullName)
        {
            return new DelegatedTestMethod(
                string.Format(CultureInfo.InvariantCulture, "Assignment_{0}_{1}", dataMember.Name, objectName), 
                () =>
                    {
                        AssertUtilities.WriteLine(String.Format(CultureInfo.InvariantCulture, "Expected exception type full name: {0}.", 
                            expectedExceptionTypeFullName));
                        AssertUtilities.WriteLine("Creating value...");
                        object value = func();
                        // ReSharper disable once ConvertIfStatementToConditionalTernaryExpression
                        if (value == null)
                            AssertUtilities.WriteLine("Value = null");
                        else
                        {
                            AssertUtilities.WriteLine(String.Format(CultureInfo.CurrentCulture, "Value of type: {0}", value.GetType()));
                            string s;
                            try
                            {
                                s = value.ToString();
                            }
                            catch (Exception)
                            {
                                s = "*** Exception in ToString()";
                            }
                            AssertUtilities.WriteLine(String.Format(CultureInfo.CurrentCulture, "Value = {0}", s));
                        }
                        if (dataMember.IsStatic())
                        {
#if (NETFRAMEWORK)
                            using (var isolatedApp = new IsolatedApp<Worker>())
                                isolatedApp.Value.Test(typeUnderTest, dataMember, value, expectedExceptionTypeFullName);
#else
                            AssertUtilities.Inconclusive("Testing static members is not supported under .NET Standard.");
#endif
                        }
                        else
                            new Worker().Test(typeUnderTest, dataMember, value, expectedExceptionTypeFullName);
                    });
        }

        protected override void Initialize(Type typeUnderTest)
        {
            base.Initialize(typeUnderTest);

            IEnumerable<MemberInfo> dataMembers = TypeUtilities.HighestLevelMembers(typeUnderTest, typeUnderTest.GetDataMembers());
            foreach (MemberInfo dataMember in dataMembers)
            {
                Debug.Assert(dataMember != null);
                if (!ShouldTestDataMember(dataMember))
                    continue;

                Type dataType = dataMember.GetDataType();
                Debug.Assert(dataType != null);

                var singularityCreators = new Dictionary<string, CreatorEntry<InvalidValueAttribute>>();
                IEnumerable<InvalidValueAttribute> singularityAttributes = 
                    dataMember.GetCustomAttributesNotNull<InvalidValueAttribute>(inherit: true);
                int i = 1;
                foreach (InvalidValueAttribute invalidValueAttribute in singularityAttributes)
                {
                    Debug.Assert(invalidValueAttribute != null);
                    string name = string.Format(CultureInfo.InvariantCulture, "_{0}", i);
                    InvalidValueAttribute attribute = invalidValueAttribute;
                    singularityCreators[name] = new CreatorEntry<InvalidValueAttribute>(
                        () => attribute.Value, invalidValueAttribute);
                    i++;
                }

                foreach (KeyValuePair<string, CreatorEntry<InvalidValueAttribute>> creator in singularityCreators)
                {
                    string objectName = creator.Key;
                    CreatorEntry<InvalidValueAttribute> creatorEntry = creator.Value;
                    Debug.Assert(objectName != null);
                    Debug.Assert(creatorEntry != null);

                    AddTestMethod(CreateAssignmentTestMethod(typeUnderTest, dataMember, objectName, creatorEntry.Creator, creatorEntry.Attribute.ExceptionTypeName));
                }

                IEnumerable<InvalidValuesAttribute> pluralityAttributes = 
                    dataMember.GetCustomAttributesNotNull<InvalidValuesAttribute>(inherit: false);
                Dictionary<string, CreatorEntry<InvalidValuesAttribute>> pluralityCreators =
                    GetCreatorEntriesFromAttributes(dataType, pluralityAttributes, allowSubclasses: true);

                foreach (KeyValuePair<string, CreatorEntry<InvalidValuesAttribute>> creator in pluralityCreators)
                {
                    string objectName = creator.Key;
                    CreatorEntry<InvalidValuesAttribute> creatorEntry = creator.Value;
                    Debug.Assert(objectName != null);
                    Debug.Assert(creatorEntry != null);

                    AddTestMethod(CreateAssignmentTestMethod(typeUnderTest, dataMember, objectName, creatorEntry.Creator, creatorEntry.Attribute.ExceptionTypeName));
                }
            }
        }

        protected virtual bool ShouldTestDataMember([NotNull] MemberInfo dataMember)
        {
            Debug.Assert(dataMember != null);
            // ReSharper disable once ConvertIfStatementToReturnStatement
            if (!dataMember.IsSettable())
                return false;

            switch (dataMember)
            {
                case FieldInfo fieldInfo:
                    return fieldInfo.IsPublicSurface();
                case PropertyInfo propertyInfo:
                    return propertyInfo.HasSetMethod() && propertyInfo.IsSetMethodPublicSurface();
            }
            return false;
        }
    }
}
