﻿
using System;
using JetBrains.Annotations;
using Microsoft.VisualStudio.TestTools.UnitTesting;
using OpcLabs.BaseLib.DataTypeModel;
using OpcLabs.BaseLibFactory.DataTypeModel;
using OpcLabs.UnitTestLib;

namespace OpcLabs.BaseLibTest.DataTypeModel
{
    public partial class StructuredDataTypeTest
    {
        #region Static Private Properties

        [NotNull]
        static private StructuredDataType Circular1
        {
            get
            {
                var circular1 = new StructuredDataType("Circular1");
                circular1.DataFields.Add(new DataField("DataField1", new PrimitiveDataType(typeof(int))));
                circular1.DataFields.Add(new DataField("DataField2", circular1));
                circular1.DataFields.Add(new DataField("DataField3", new PrimitiveDataType(typeof(int))));
                return circular1;
            }
        }

        [NotNull]
        static private StructuredDataType Circular2
        {
            get
            {
                var intermediary = new StructuredDataType("Intermediary");

                var circular2 = new StructuredDataType("Circular2");
                circular2.DataFields.Add(new DataField("Cdf1", new PrimitiveDataType(typeof(int))));
                circular2.DataFields.Add(new DataField("Cdf2", intermediary));
                circular2.DataFields.Add(new DataField("Cdf3", new PrimitiveDataType(typeof(int))));

                intermediary.DataFields.Add(new DataField("Idf1", new PrimitiveDataType(typeof(int))));
                intermediary.DataFields.Add(new DataField("Idf2", circular2));
                intermediary.DataFields.Add(new DataField("Idf3", new PrimitiveDataType(typeof(int))));

                return circular2;
            }
        }

        [NotNull]
        static private StructuredDataType Null => new StructuredDataType();

        #endregion



        #region Test Methods

        [TestMethod]
        public void EqualsTest_Circular1_Circular2()
        {
            bool b = Circular1.Equals(Circular2);
            Assert.IsFalse(b);
        }

        [ExpectedException(typeof(ArgumentException))]
        [TestMethod]
        public void EqualsTest_Circular1_DifferentInstances()
        {
            // ReSharper disable once EqualExpressionComparison
            // ReSharper disable once ReturnValueOfPureMethodIsNotUsed
            Circular1.Equals(Circular1);
        }

        [TestMethod]
        public void EqualsTest_Circular1_SameInstances()
        {
            StructuredDataType instance = Circular1;
            // ReSharper disable once EqualExpressionComparison
            bool b = instance.Equals(instance);
            Assert.IsTrue(b);
        }

        [TestMethod]
        public void EqualsTest_Circular2_Circular1()
        {
            bool b = Circular2.Equals(Circular1);
            Assert.IsFalse(b);
        }

        [ExpectedException(typeof(ArgumentException))]
        [TestMethod]
        public void EqualsTest_Circular2_DifferentInstances()
        {
            // ReSharper disable once EqualExpressionComparison
            // ReSharper disable once ReturnValueOfPureMethodIsNotUsed
            Circular2.Equals(Circular2);
        }

        [TestMethod]
        public void EqualsTest_Circular2_SameInstances()
        {
            StructuredDataType instance = Circular2;
            // ReSharper disable once EqualExpressionComparison
            bool b = instance.Equals(instance);
            Assert.IsTrue(b);
        }

        [TestMethod]
        public void GetHashCodeTest_Circular1()
        {
            // ReSharper disable once ReturnValueOfPureMethodIsNotUsed
            Circular1.GetHashCode();
            // must not cause exception due to cycles
        }

        [TestMethod]
        public void GetHashCodeTest_Circular2()
        {
            // ReSharper disable once ReturnValueOfPureMethodIsNotUsed
            Circular2.GetHashCode();
            // must not cause exception due to cycles
        }

        [TestMethod]
        public void IsCompleteTest_Negatives()
        {
            Assert.IsFalse(StructuredDataTypeInstances.IncompleteInstance1.IsComplete);
            Assert.IsFalse(StructuredDataTypeInstances.IncompleteInstance2.IsComplete);
        }

        [TestMethod]
        public void IsCompleteTest_Positives()
        {
            Assert.IsTrue(Circular1.IsComplete);
            Assert.IsTrue(Circular2.IsComplete);
            Assert.IsTrue(Null.IsComplete);
            Assert.IsTrue(StructuredDataTypeInstances.Instance0.IsComplete);
            Assert.IsTrue(StructuredDataTypeInstances.Instance1.IsComplete);
            Assert.IsTrue(StructuredDataTypeInstances.Instance2.IsComplete);
            Assert.IsTrue(StructuredDataTypeInstances.Instance3.IsComplete);
            Assert.IsTrue(StructuredDataTypeInstances.Instance4.IsComplete);
            Assert.IsTrue(StructuredDataTypeInstances.Instance5.IsComplete);
        }

        [TestMethod]
        public void IsCyclicTest_Circular1()
        {
            Assert.IsTrue(Circular1.IsRecursive);
        }

        [TestMethod]
        public void IsCyclicTest_Circular2()
        {
            Assert.IsTrue(Circular2.IsRecursive);
        }

        [TestMethod]
        public void IsCyclicTest_Null()
        {
            Assert.IsFalse(Null.IsRecursive);
        }

        [TestMethod]
        public void IsTerminableTest_Negatives()
        {
            Assert.IsFalse(Circular1.IsTerminable);
            Assert.IsFalse(Circular2.IsTerminable);
        }

        [TestMethod]
        public void IsTerminableTest_Positives()
        {
            Assert.IsTrue(Null.IsTerminable);
            Assert.IsTrue(StructuredDataTypeInstances.IncompleteInstance1.IsTerminable);
            Assert.IsTrue(StructuredDataTypeInstances.IncompleteInstance2.IsTerminable);
            Assert.IsTrue(StructuredDataTypeInstances.Instance0.IsTerminable);
            Assert.IsTrue(StructuredDataTypeInstances.Instance1.IsTerminable);
            Assert.IsTrue(StructuredDataTypeInstances.Instance2.IsTerminable);
            Assert.IsTrue(StructuredDataTypeInstances.Instance3.IsTerminable);
            Assert.IsTrue(StructuredDataTypeInstances.Instance4.IsTerminable);
            Assert.IsTrue(StructuredDataTypeInstances.Instance5.IsTerminable);
        }

        [TestMethod]
        public void ToStringTest_Circular1()
        {
            StructuredDataType target = Circular1;
            string s = target.ToString();
            Assert.IsNotNull(s);
            AssertUtilities.WriteLine(s);
            Assert.AreEqual("Circular1 = structured", s);
        }

        [TestMethod]
        public void ToStringTest_Circular1_Verbose()
        {
            StructuredDataType target = Circular1;
            string s = target.ToString("V", null);
            AssertUtilities.WriteLine(s);
            Assert.IsNotNull(s);
            // IMPROVE: check the output
        }

        [TestMethod]
        public void ToStringTest_Circular2()
        {
            StructuredDataType target = Circular2;
            string s = target.ToString();
            Assert.IsNotNull(s);
            AssertUtilities.WriteLine(s);
            Assert.AreEqual("Circular2 = structured", s);
        }

        [TestMethod]
        public void ToStringTest_Circular2_Verbose()
        {
            StructuredDataType target = Circular2;
            string s = target.ToString("V", null);
            Assert.IsNotNull(s);
            AssertUtilities.WriteLine(s);
            // IMPROVE: check the output
        }

        [TestMethod]
        public void ToStringTest_Null()
        {
            StructuredDataType target = Null;
            string s = target.ToString();
            Assert.IsNotNull(s);
            AssertUtilities.WriteLine(s);
            Assert.AreEqual("structured", s);
        }

        [TestMethod]
        public void ToStringTest_Null_Verbose()
        {
            StructuredDataType target = Null;
            string s = target.ToString("V", null);
            AssertUtilities.WriteLine(s);
            Assert.IsNotNull(s);
            Assert.AreEqual("structured" + Environment.NewLine, s);
        }

        #endregion
    }
}
