﻿using System;
using System.Collections.Generic;
using System.Net.NetworkInformation;
using JetBrains.Annotations;
using Microsoft.VisualStudio.TestTools.UnitTesting;
using OpcLabs.BaseLib.Collections.Generic.Extensions;
using OpcLabs.BaseLib.Extensions.Internal;
using OpcLabs.BaseLib.Net.NetworkInformation;
using OpcLabs.BaseLib.Net.Sockets;
using OpcLabs.BaseLib.Networking.Internal;
using OpcLabs.EasyOpc;

namespace OpcLabs.EasyOpcLiveTest.Legacy
{
    class LegacyTestEnvironment
    {
        private LegacyTestEnvironment()
        {
        }

        [NotNull]
        static public readonly LegacyTestEnvironment Current = new LegacyTestEnvironment();



        #region Public Methods

#if (NETFRAMEWORK)
        public void Precheck([NotNull] ServerDescriptor serverDescriptor)
        {
            string message;
            lock (_precheckCache)
            {
                if (!_precheckCache.TryGetValue(serverDescriptor, out message))
                {
                    message = InternalPrecheck(serverDescriptor);
                    _precheckCache.AddNew(serverDescriptor, message);
                }
            }
            if (message != null)
                Assert.Inconclusive(message);
        }
#endif

#endregion



#region Private Methods

#if (NETFRAMEWORK)
        [CanBeNull]
        protected string InternalPrecheck([NotNull] ServerDescriptor serverDescriptor)
        {
            // IMPROVE: have this for !(NETFRAMEWORK)
#if (NETFRAMEWORK)
            switch (serverDescriptor.Technology)
            {
                case OpcTechnologies.Com:
                    {
                        string machineName = serverDescriptor.MachineName;
                        if (!machineName.IsEmpty())
                        {
                            bool success;
                            PingReply pingReply = null;
                            try
                            {
                                pingReply = PingUtilities.Send(machineName);
                                success = true;
                            }
                            catch (NotSupportedException)
                            {
                                // address is an IPv6 address and the local computer is running an operating system earlier than Windows 2000. 
                                success = false;
                            }
                            catch (PingException)
                            {
                                // An exception was thrown while sending or receiving the ICMP messages. See the inner exception for the exact 
                                // exception that was thrown.
                                success = false;
                            }
                            if (!success || (pingReply.Status != IPStatus.Success))
                                return $"Failed pre-check: Ping test failure (\"{machineName}\").";
                        }

                        // ServerDescriptor pre-check passes for all pingable appear-remote machines, regardless of the server class.
                        bool appearsLocal =
                            machineName.IsEmpty() ||
                            HostNameUtilities.IsLocalHostOrAddress(machineName) ||
                            (machineName.ToUpper() == System.Environment.MachineName.ToUpper()) ||
                            HostNameUtilities.Match(machineName, System.Net.Dns.GetHostName());
                        // Cannot check instantiation feasibility on remote machines.
                        if (appearsLocal)
                        {
                            ServerDescriptor localServerDescriptor = serverDescriptor.TypedClone();
                            localServerDescriptor.MachineName = "";
                            if (!localServerDescriptor.CanInstantiate())
                                return $"Failed pre-check: Cannot instantiate (\"{localServerDescriptor}\").";
                        }

                        break;
                    }
                case OpcTechnologies.Xml:
                    {
                        if (!TcpClientUtilities.CanCreate(serverDescriptor))
                            return $"Failed pre-check: Socket exception (\"{serverDescriptor.Host}\", {serverDescriptor.Port}).";
                        break;
                    }
            }
#endif
            return null;
        }
#endif

#endregion



#region Private Fields

#if (NETFRAMEWORK)
        [NotNull]
        private readonly Dictionary<ServerDescriptor, string> _precheckCache = new Dictionary<ServerDescriptor, string>();
#endif

#endregion
    }
}
