﻿using System;
using System.Collections.Generic;
using System.Diagnostics;
using JetBrains.Annotations;
using Microsoft.VisualStudio.TestTools.UnitTesting;
using OpcLabs.BaseLib.Abstract;
using OpcLabs.BaseLib.Connectivity.Point;
using OpcLabs.BaseLib.Connectivity.Point.Reactive;
using OpcLabs.BaseLib.Data;
using OpcLabs.BaseLib.Extensions;
using OpcLabs.EasyOpc.UA;
using OpcLabs.EasyOpc.UA.Connectivity;
using OpcLabs.EasyOpc.UA.Connectivity.Implementation;
using OpcLabs.EasyOpcLiveTest.UA.Environment;
using OpcLabs.UnitTestLib;

namespace OpcLabs.EasyOpcLiveTest.UA.Connectivity
{
    [TestClass]
    public class UAConnectivityTest
    {
        #region Static Private Fields

        static private int _subscribeInvocationId;

        #endregion



        #region Test Methods

        [TestCategory("Live")]
        [TestCategory("OpcUA")]
        [TestMethod]
        public void CreateReaderAndReadAttributePointTest()
        {
            TestSection.Arrange();
            UAEndpointDescriptor endpointDescriptor = UATestEnvironment.OpcFoundation_UASampleServer.EndpointDescriptors.OpcTcp;
            TestEnvironment.PrecheckEndpoint(endpointDescriptor);

            var connectivity = new UAConnectivity();

            // ReSharper disable once PossibleNullReferenceException
            var pointReader = (UAPointReader) connectivity.GetService<IReadPointService>().CreateReader();
            Assert.IsNotNull(pointReader);
            Assert.IsFalse(pointReader.IsFrozen);

            var itemPoint = new UAAttributePoint(endpointDescriptor, "nsu=http://test.org/UA/Data/;ns=2;i=10853");
            var parameters = new UAAttributePointReadParameters();
            var arguments = new UAAttributePointReadArguments();

            TestSection.Act();
            pointReader.Read(itemPoint, parameters, arguments);

            TestSection.Assert();

            AssertUtilities.WriteLine(arguments.Exception);
            Assert.IsNull(arguments.Exception);
            Assert.AreEqual(0, arguments.ErrorCode);
            Assert.AreEqual("", arguments.ErrorMessage);
            Assert.AreEqual(StatusInfo.Normal, arguments.StatusInfo);
            // ReSharper disable once RedundantCast
            Assert.AreEqual((UAStatusCode)UASeverity.GoodOrSuccess, arguments.StatusCode);
            object value = arguments.Value;
            AssertUtilities.WriteLine(value);
            Assert.IsNotNull(value);
            Assert.IsInstanceOfType(value, typeof(Single));
        }

        [TestCategory("Live")]
        [TestCategory("OpcUA")]
        [TestCategory("TimeDomain")]
        [TestMethod]
        public void CreateSubscriberAndSubscribeAttributePointTest()
        {
            var connectivity = new UAConnectivity();
            CreateSubscriberAndSubscribeAttributePointTest(connectivity, subscribeDuration:60*1000, delayAfterUnsubscribe:5*1000);
        }

        [TestCategory("Live")]
        [TestCategory("OpcUA")]
        [TestMethod]
        public void CreateWriterAndWriteAttributePointTest()
        {
            TestSection.Arrange();
            UAEndpointDescriptor endpointDescriptor = UATestEnvironment.OpcFoundation_UASampleServer.EndpointDescriptors.OpcTcp;
            TestEnvironment.PrecheckEndpoint(endpointDescriptor);

            var connectivity = new UAConnectivity();
            // ReSharper disable once PossibleNullReferenceException
            var pointWriter = (UAPointWriter)connectivity.GetService<IWritePointService>().CreateWriter();
            Assert.IsNotNull(pointWriter);
            Assert.IsFalse(pointWriter.IsFrozen);

            var itemPoint = new UAAttributePoint(endpointDescriptor, OpcFoundation_UASampleServer.NodeIdExpandedTexts.Data_Static_UserScalar_Int32Value);
            var parameters = new UAAttributePointWriteParameters();
            var arguments = new UAAttributePointWriteArguments {Value = 123456};

            TestSection.Act();
            pointWriter.Write(itemPoint, parameters, arguments);

            TestSection.Assert();

            AssertUtilities.WriteLine($"Exception: {arguments.Exception}");
            Assert.IsNull(arguments.Exception);
            Assert.AreEqual(0, arguments.ErrorCode);
            Assert.AreEqual("", arguments.ErrorMessage);
            Assert.AreEqual(StatusInfo.Normal, arguments.StatusInfo);
            // ReSharper disable once RedundantCast
            Assert.AreEqual((UAStatusCode)UASeverity.GoodOrSuccess, arguments.StatusCode);
            Assert.IsFalse(arguments.Clamped);
            Assert.IsFalse(arguments.CompletesAsynchronously);
        }

        [TestCategory("Live")]
        [TestCategory("OpcUA")]
        [TestCategory("TimeDomain")]
        [TestMethod]
        public void ResubscribeTest_AfterHoldPeriod()
        {
            int delayAfterUnsubscribe = EasyUAClient.AdaptableParameters.SessionParameters.HoldPeriod + 5*1000;
            ResubscribeTest(delayAfterUnsubscribe);
        }

        [TestCategory("Live")]
        [TestCategory("OpcUA")]
        [TestCategory("TimeDomain")]
        [TestMethod]
        public void ResubscribeTest_BeforeHoldPeriod()
        {
            int delayAfterUnsubscribe = (int)(EasyUAClient.AdaptableParameters.SessionParameters.HoldPeriod / 2.0);
            ResubscribeTest(delayAfterUnsubscribe);
        }

        [TestCategory("Live")]
        [TestCategory("OpcUA")]
        [TestCategory("TimeDomain")]
        [TestMethod]
        public void SubscribePointReactiveTest()
        {
            const int waitMilliseconds = 15*1000;
            const int spinUpReserve = 8*1000;
            const int spinDownWait = 5*1000;

            TestSection.Arrange();
            UAEndpointDescriptor endpointDescriptor = UATestEnvironment.OpcFoundation_UASampleServer.EndpointDescriptors.OpcTcp;
            TestEnvironment.PrecheckEndpoint(endpointDescriptor);

            var connectivity = new UAConnectivity();
            ISubscribePointService subscribePointService = connectivity.GetService<ISubscribePointService>();
            Debug.Assert(subscribePointService != null);
            var subscribePointReactive = new SubscribePointReactive(subscribePointService);

            var point = new UAAttributePoint(endpointDescriptor, OpcFoundation_UASampleServer.NodeIdExpandedTexts.Data_Dynamic_UserScalar_Int32Value);
            var parameters = new UAAttributePointSubscribeParameters();
            object SuccessSelector(object o) => o;
            object FailureSelector(object o) => o;
            PointChangedObservable pointChangedObservable = subscribePointReactive
                .CreatePointChangedObservable(
                    point, 
                    parameters, 
                    SuccessSelector, FailureSelector);

            TestSection.Act();

            int nextCount = 0;
            using (pointChangedObservable.Subscribe(arg =>
                {
                    nextCount++;
                    AssertUtilities.WriteLine(arg);
                }))
            {
                AssertUtilities.WaitingPhase(waitMilliseconds);
            }
            AssertUtilities.WaitingPhase(spinDownWait);
            AssertUtilities.WriteLine($"Next count: {nextCount}");

            TestSection.Assert();

            int minimumNextCount = (waitMilliseconds - spinUpReserve)/parameters.MonitoringParameters.SamplingInterval;
            AssertUtilities.WriteLine($"Minimum Next count: {minimumNextCount}");
            Assert.IsTrue(nextCount >= minimumNextCount);
            // IMPROVE: have more asserts
        }

        #endregion



        #region Private Methods

        private void CreateSubscriberAndSubscribeAttributePointTest(
            [NotNull] UAConnectivity connectivity,
            int subscribeDuration,
            int delayAfterUnsubscribe)
        {
            TestSection.Arrange();
            UAEndpointDescriptor endpointDescriptor = UATestEnvironment.OpcFoundation_UASampleServer.EndpointDescriptors.OpcTcp;
            TestEnvironment.PrecheckEndpoint(endpointDescriptor);

            _subscribeInvocationId++;
            AssertUtilities.WriteLine($"New subscribe invocation ID: {_subscribeInvocationId}");

            AssertUtilities.Phase("Creating subscriber");
            // ReSharper disable once PossibleNullReferenceException
            var pointSubscriber = (UAPointSubscriber) connectivity.GetService<ISubscribePointService>().CreateSubscriber();
            Assert.IsNotNull(pointSubscriber);
            Assert.IsFalse(pointSubscriber.IsFrozen);

            var state = new Wrapper<int>(_subscribeInvocationId);
            var attributePoint = new UAAttributePoint(endpointDescriptor, "nsu=http://test.org/UA/Data/;ns=2;i=10853");
            var parameters = new UAAttributePointSubscribeParameters();
            var arguments = new UAAttributePointSubscribeArguments();

            int pointChangedCount = 0;
            var assertFailedExceptionList = new List<AssertFailedException>();
            pointSubscriber.PointChanged += (sender, args) =>
            {
                try
                {
                    AssertUtilities.WriteLine($"Point changed, subscribe invocation ID: {_subscribeInvocationId}");

                    pointChangedCount++;

                    Assert.AreSame(pointSubscriber, sender);

                    Assert.IsNotNull(args);
                    Assert.AreSame(state, args.State, "State differs");
                    Assert.AreSame(attributePoint, args.Point, "Point differs");
                    Assert.AreSame(parameters, args.Parameters, "Parameters differ");
                    Assert.AreSame(arguments, args.Arguments, "Arguments differ");

                    AssertUtilities.WriteLine($"Exception: {arguments.Exception}");
                    Assert.IsNull(arguments.Exception);
                    Assert.AreEqual(0, arguments.ErrorCode);
                    Assert.AreEqual("", arguments.ErrorMessage);
                    Assert.AreEqual(StatusInfo.Normal, arguments.StatusInfo);
                    // ReSharper disable once RedundantCast
                    Assert.AreEqual((UAStatusCode)UASeverity.GoodOrSuccess, arguments.StatusCode);
                    object value = arguments.Value;
                    AssertUtilities.WriteLine($"Value: {value}");
                    Assert.IsNotNull(value);
                    Assert.IsInstanceOfType(value, typeof(Single));
                }
                catch (AssertFailedException assertFailedException)
                {
                    assertFailedExceptionList.Add(assertFailedException);
                }
            };

            TestSection.Act();

            AssertUtilities.Phase("Subscribing");
            object handle = pointSubscriber.Subscribe(state, attributePoint, parameters, arguments);
            AssertUtilities.WaitingPhase(subscribeDuration);

            AssertUtilities.Phase("Unsubscribing");
            pointSubscriber.Unsubscribe(handle);
            AssertUtilities.WaitingPhase(delayAfterUnsubscribe);

            TestSection.Assert();

            AssertUtilities.WriteLine($"Assert failed exceptions: {assertFailedExceptionList.Count}");
            AssertUtilities.WriteLine($"Point changed count: {pointChangedCount}");
            Assert.IsTrue(pointChangedCount >= 1, "Not enough point changes");

            foreach (AssertFailedException assertFailedException in assertFailedExceptionList)
            {
                Debug.Assert(assertFailedException != null);
                throw assertFailedException;
            }
        }
        public void ResubscribeTest(int delayAfterUnsubscribe)
        {
            const int loopCount = 5;
            const int subscribeDuration = 10 * 1000;

            AssertUtilities.WriteLine($"Delay after Unsubscribe: {delayAfterUnsubscribe}");

            var connectivity = new UAConnectivity();
            for (int iteration = 1; iteration <= loopCount; iteration++)
            {
                AssertUtilities.WriteLine($"Iteration: {iteration}");
                CreateSubscriberAndSubscribeAttributePointTest(connectivity, subscribeDuration, delayAfterUnsubscribe);
            }
        }

        #endregion



        #region Private Properties

        [NotNull]
        private UATestEnvironment TestEnvironment => UATestEnvironment.Current;

        #endregion
    }
}
