﻿using System.Collections.Generic;
using Microsoft.VisualStudio.TestTools.UnitTesting;
using OpcLabs.BaseLib.OperationModel;
using OpcLabs.EasyOpc.UA;
using OpcLabs.EasyOpc.UA.AddressSpace;
using OpcLabs.EasyOpc.UA.AddressSpace.Standard;
using OpcLabs.EasyOpc.UA.AlarmsAndConditions;
using OpcLabs.EasyOpc.UA.Filtering;
using OpcLabs.EasyOpcLiveTest.UA.Environment;
using OpcLabs.UnitTestLib;

namespace OpcLabs.EasyOpcLiveTest.UA
{
    public partial class EasyUAClientTest
    {
        #region Test Methods

        [TestCategory("Live")]
        [TestCategory("OpcUA")]
        [TestCategory("TimeDomain")]
        [TestCategory("UAAlarmsAndConditions")]
        [TestMethod]
        public void EventNotificationTest_Basic()
        {
            TestSection.Arrange();
            UAEndpointDescriptor endpointDescriptor = UATestEnvironment.OpcFoundation_UAAlarmConditionServer.EndpointDescriptors.OpcTcp;
            TestEnvironment.PrecheckEndpoint(endpointDescriptor);

            var target = new EasyUAClient();

            int failureCount = 0;
            int successCount = 0;
            int dataCount = 0;
            target.EventNotification += (sender, args) =>
            {
                Assert.IsNotNull(args);
                AssertUtilities.WriteLine(args);
                if (args.Succeeded)
                {
                    successCount++;
                    UAEventData eventData = args.EventData;
                    if (eventData != null)
                    {
                        dataCount++;
                        foreach (KeyValuePair<UAAttributeField, ValueResult> pair in eventData.FieldResults)
                            AssertUtilities.WriteLine($"{pair.Key} -> {pair.Value}");
                    }
                }
                else
                    failureCount++;
            };

            TestSection.Act();

            int handle = target.SubscribeEvent(
                endpointDescriptor,
                UAObjectIds.Server,
                1000);

            AssertUtilities.WaitingPhase(30*1000);

            target.UnsubscribeMonitoredItem(handle);

            TestSection.Assert();

            AssertUtilities.WriteLine($"Failure count: {failureCount}");
            AssertUtilities.WriteLine($"Success count: {successCount}");
            AssertUtilities.WriteLine($"Data count: {dataCount}");

            Assert.IsTrue(dataCount >= 1, "no data notifications");
            Assert.IsTrue(successCount >= 2, "not enough successful notifications");
            Assert.AreEqual(0, failureCount, "some failed notifications");
        }

        [TestCategory("Live")]
        [TestCategory("OpcUA")]
        [TestCategory("TimeDomain")]
        [TestCategory("UAAlarmsAndConditions")]
        [TestMethod]
        public void EventNotificationTest_WhereClause_Or()
        {
            TestSection.Arrange();
            UAEndpointDescriptor endpointDescriptor = UATestEnvironment.OpcFoundation_UAAlarmConditionServer.EndpointDescriptors.OpcTcp;
            TestEnvironment.PrecheckEndpoint(endpointDescriptor);

            const int minimumSeverity = 500;
            var requiredSourceNode = new UANodeId(expandedText:"nsu=http://opcfoundation.org/Quickstarts/AlarmCondition;ns=2;s=1:Metals/SouthMotor");

            var target = new EasyUAClient();

            UAEventFilter eventFilter = new UAEventFilterBuilder(
                UAFilterElements.Or(
                    UAFilterElements.GreaterThanOrEqual(UABaseEventObject.Operands.Severity, minimumSeverity),
                    UAFilterElements.Equals(UABaseEventObject.Operands.SourceNode, requiredSourceNode)),
                UABaseEventObject.AllFields);

            int failureCount = 0;
            int successCount = 0;
            int violationCount = 0;
            int dataCount = 0;
            target.EventNotification += (sender, args) =>
            {
                Assert.IsNotNull(args);
                AssertUtilities.WriteLine(args);
                if (args.Succeeded)
                {
                    successCount++;
                    UAEventData eventData = args.EventData;
                    if (eventData != null)
                    {
                        dataCount++;
                        foreach (KeyValuePair<UAAttributeField, ValueResult> pair in eventData.FieldResults)
                            AssertUtilities.WriteLine($"{pair.Key} -> {pair.Value}");

                        int severity = eventData.BaseEvent.Severity;
                        UANodeId sourceNode = eventData.BaseEvent.SourceNode;
                        bool orOperand1Result = severity >= minimumSeverity;
                        bool orOperand2Result = requiredSourceNode.Match(sourceNode);
                        bool orResult = orOperand1Result || orOperand2Result;
                        if (!orResult)
                        {
                            AssertUtilities.WriteLine("Where clause violation");
                            violationCount++;
                        }
                    }
                }
                else
                    failureCount++;
            };

            TestSection.Act();
            int handle = target.SubscribeEvent(
                endpointDescriptor,
                UAObjectIds.Server,
                1000,
                eventFilter);

            AssertUtilities.WaitingPhase(30 * 1000);

            target.UnsubscribeMonitoredItem(handle);

            TestSection.Assert();

            AssertUtilities.WriteLine($"Data count: {dataCount}");
            AssertUtilities.WriteLine($"Success count: {successCount}");
            AssertUtilities.WriteLine($"Failure count: {failureCount}");
            AssertUtilities.WriteLine($"Violation count: {violationCount}");

            Assert.IsTrue(dataCount >= 1, "no data notifications");
            Assert.IsTrue(successCount >= 2, "not enough successful notifications");
            Assert.AreEqual(0, failureCount, "some failed notifications");
            Assert.AreEqual(0, violationCount, "some filter violations");
        }

        [TestCategory("Live")]
        [TestCategory("OpcUA")]
        [TestCategory("TimeDomain")]
        [TestCategory("UAAlarmsAndConditions")]
        [TestMethod]
        public void EventNotificationTest_WhereClause_Severity()
        {
            TestSection.Arrange();
            UAEndpointDescriptor endpointDescriptor = UATestEnvironment.OpcFoundation_UAAlarmConditionServer.EndpointDescriptors.OpcTcp;
            TestEnvironment.PrecheckEndpoint(endpointDescriptor);

            const int minimumSeverity = 500;

            var target = new EasyUAClient();

            UAEventFilter eventFilter = new UAEventFilterBuilder(
                UAFilterElements.GreaterThanOrEqual(
                    UABaseEventObject.Operands.Severity,    //UAFilterElements.SimpleAttribute(UAObjectTypeIds.BaseEventType, "/Severity"),
                    minimumSeverity),   //UAFilterElements.Literal(minimumSeverity)
                UABaseEventObject.AllFields);

            int failureCount = 0;
            int successCount = 0;
            int violationCount = 0;
            int dataCount = 0;
            target.EventNotification += (sender, args) =>
            {
                Assert.IsNotNull(args);
                AssertUtilities.WriteLine(args);
                if (args.Succeeded)
                {
                    successCount++;
                    UAEventData eventData = args.EventData;
                    if (eventData != null)
                    {
                        dataCount++;
                        foreach (KeyValuePair<UAAttributeField, ValueResult> pair in eventData.FieldResults)
                            AssertUtilities.WriteLine($"{pair.Key} -> {pair.Value}");

                        int severity = eventData.BaseEvent.Severity;
                        if (severity < minimumSeverity)
                        {
                            AssertUtilities.WriteLine(
                                $"Severity {severity} is less than the minimum {minimumSeverity} specified in the filter");
                            violationCount++;
                        }
                    }
                }
                else
                    failureCount++;
            };

            TestSection.Act();

            int handle = target.SubscribeEvent(
                endpointDescriptor,
                UAObjectIds.Server,
                1000,
                eventFilter);

            AssertUtilities.WaitingPhase(30 * 1000);

            target.UnsubscribeMonitoredItem(handle);

            TestSection.Assert();

            AssertUtilities.WriteLine($"Data count: {dataCount}");
            AssertUtilities.WriteLine($"Success count: {successCount}");
            AssertUtilities.WriteLine($"Failure count: {failureCount}");
            AssertUtilities.WriteLine($"Violation count: {violationCount}");

            Assert.IsTrue(dataCount >= 1, "no data notifications");
            Assert.IsTrue(successCount >= 2, "not enough successful notifications");
            Assert.AreEqual(0, failureCount, "some failed notifications");
            Assert.AreEqual(0, violationCount, "some filter violations");
        }

        #endregion
    }
}
