﻿
using System;
using System.Diagnostics;
using System.Linq;
using JetBrains.Annotations;
using Microsoft.VisualStudio.TestTools.UnitTesting;
using OpcLabs.BaseLib.DataTypeModel;
using OpcLabs.BaseLib.OperationModel;
using OpcLabs.BaseLib.OperationModel.Generic;
using OpcLabs.EasyOpc.UA.DataTypeModel;
using OpcLabs.EasyOpc.UA.DataTypeModel.Implementation;
using OpcLabs.EasyOpc.UA.InformationModel;
using OpcLabs.UnitTestLib;

namespace OpcLabs.EasyOpcTest.UA.DataTypeModel.Internal
{
    public class UAOpcBinaryDataTypeSystemTest
    {
        #region Static Internal Methods

        static internal void DecodeErrorTest(
            [CanBeNull] Action<UAOpcBinaryDts> setupTargetAction,
            [NotNull] byte[] dictionaryByteArray,
            [CanBeNull] object dataTypeDescription,
            [NotNull] byte[] extensionObjectBody,
            int expectedErrorCode)
        {
            ValueResult<GenericData> decodedValueResult =
                DecodeTestCommon(setupTargetAction, dictionaryByteArray, dataTypeDescription, extensionObjectBody);
            Assert.IsFalse(decodedValueResult.Succeeded, "Decoding has succeeded but should have failed.");
            Assert.AreEqual(expectedErrorCode, decodedValueResult.ErrorCode, "Error codes differ.");
        }

        [NotNull]
        static internal GenericData DecodeTest(
            [CanBeNull] Action<UAOpcBinaryDts> setupTargetAction,
            [NotNull] byte[] dictionaryByteArray,
            [CanBeNull] object dataTypeDescription,
            [NotNull] byte[] extensionObjectBody)
        {
            ValueResult<GenericData> decodedValueResult =
                DecodeTestCommon(setupTargetAction, dictionaryByteArray, dataTypeDescription, extensionObjectBody);
            Assert.IsTrue(decodedValueResult.Succeeded, "Decoding has failed but should have succeeded.");
            GenericData decodedValue = decodedValueResult.Value;
            AssertUtilities.WriteLine($"decodedValue: {decodedValue:V}");
            Assert.IsNotNull(decodedValue, "Decoded value is a null reference.");
            Exception validationException = decodedValue.TryValidate();
            AssertUtilities.WriteLine($"validationException: {validationException}");
            Assert.IsNull(validationException, "Decoded value validation failed.");
            Assert.IsTrue(decodedValue.IsValid, "Decoded value is not valid.");
            Assert.AreEqual(0, decodedValueResult.DiagnosticsCount, "No warnings expected.");
            return decodedValue;
        }

        static internal GenericData DecodeTest(
            [CanBeNull] Action<UAOpcBinaryDts> setupTargetAction,
            [NotNull] byte[] dictionaryByteArray,
            [CanBeNull] object dataTypeDescription,
            [NotNull] byte[] extensionObjectBody,
            [NotNull] DataType expectedDataType)
        {
            GenericData decodedValue = DecodeTest(setupTargetAction, dictionaryByteArray, dataTypeDescription, extensionObjectBody);
            Assert.AreEqual(expectedDataType.Kind, decodedValue.DataTypeKind, "Data type kinds differ.");
            AssertUtilities.WriteLine($"decodedValue.DataType: {decodedValue.DataType:V}");
            AssertUtilities.AreEqualInfos(expectedDataType, decodedValue.DataType);
            return decodedValue;
        }

        static internal void DecodeTest(
            [CanBeNull] Action<UAOpcBinaryDts> setupTargetAction,
            [NotNull] byte[] dictionaryByteArray,
            [CanBeNull] object dataTypeDescription,
            [NotNull] byte[] extensionObjectBody,
            [NotNull] GenericData expectedGenericData)
        {
            DataType expectedDataType = expectedGenericData.DataType;
            GenericData decodedValue = (expectedDataType != null)
                ? DecodeTest(setupTargetAction, dictionaryByteArray, dataTypeDescription, extensionObjectBody,
                    expectedDataType)
                : DecodeTest(setupTargetAction, dictionaryByteArray, dataTypeDescription, extensionObjectBody);
            AssertUtilities.AreEqualInfos(expectedGenericData, decodedValue);
        }
        
        [NotNull]
        static internal ValueResult<GenericData> DecodeTestCommon(
            [CanBeNull] Action<UAOpcBinaryDts> setupTargetAction,
            [NotNull] byte[] dictionaryByteArray,
            [CanBeNull] object dataTypeDescription,
            [NotNull] byte[] extensionObjectBody)
        {
            TestSection.Arrange();

            var target = new UAOpcBinaryDts();
            var dataTypeDictionaryDescriptor = new UAModelNodeDescriptor();
            target.AddDataTypeDictionary(dataTypeDictionaryDescriptor, dictionaryByteArray);
            setupTargetAction?.Invoke(target);

            TestSection.Act();

            ValueResult<GenericData> decodedValueResult = target.Decode(
                new UADataTypeDescriptor(dataTypeDictionaryDescriptor, dataTypeDescription),
                extensionObjectBody);

            TestSection.Assert();

            AssertUtilities.WriteLine($"decodedValueResult: {decodedValueResult}");
            Assert.IsNotNull(decodedValueResult, "Decoded value result is a null reference.");
            AssertUtilities.WriteLine($"encodedValueResult.ErrorCode: {decodedValueResult.ErrorCode}");
            Exception exception = decodedValueResult.Exception;
            while (exception != null)
            {
                AssertUtilities.WriteLine($"exception.Message: {exception.Message}");
                exception = exception.InnerException;
            }
            AssertUtilities.WriteLine();
            return decodedValueResult;
        }

        static internal void EncodeErrorTest(
            [CanBeNull] Action<UAOpcBinaryDts> setupTargetAction,
            [NotNull] byte[] dictionaryByteArray,
            [CanBeNull] object dataTypeDescription,
            [NotNull] GenericData genericData,
            int expectedErrorCode)
        {
            Debug.Assert(expectedErrorCode <= 0);
            ValueResult encodedValueResult = EncodeTestCommon(setupTargetAction, dictionaryByteArray, dataTypeDescription, genericData);
            Assert.IsFalse(encodedValueResult.Succeeded, "Encoding has succeeded but it should have failed.");
            Assert.AreEqual(expectedErrorCode, encodedValueResult.ErrorCode, "Error codes differ.");
        }

        static internal void EncodeTest(
            [CanBeNull] Action<UAOpcBinaryDts> setupTargetAction,
            [NotNull] byte[] dictionaryByteArray,
            [CanBeNull] object dataTypeDescription,
            [NotNull] GenericData genericData,
            [NotNull] byte[] expectedExtensionObjectBody)
        {
            ValueResult encodedValueResult = EncodeTestCommon(setupTargetAction, dictionaryByteArray, dataTypeDescription, genericData);
            Assert.IsTrue(encodedValueResult.Succeeded, "Encoding has failed but it should have succeeded.");
            object encodedValue = encodedValueResult.Value;
            AssertUtilities.WriteLine($"encodedValue: {encodedValue}");
            Assert.IsNotNull(encodedValue, "Encoded value is a null reference.");
            var extensionObjectBody = encodedValue as byte[];
            Assert.IsNotNull(extensionObjectBody, "Encoded valus is not an array of bytes.");
            AssertUtilities.WriteLine($"extensionObjectBody: {BitConverter.ToString(extensionObjectBody)}");
            Assert.IsTrue(extensionObjectBody.SequenceEqual(expectedExtensionObjectBody), "Extension object body is not as expected.");
            Assert.AreEqual(0, encodedValueResult.DiagnosticsCount, "No warnings expected.");
        }

        [NotNull]
        static internal ValueResult EncodeTestCommon(
            [CanBeNull] Action<UAOpcBinaryDts> setupTargetAction,
            [NotNull] byte[] dictionaryByteArray,
            [CanBeNull] object dataTypeDescription,
            [NotNull] GenericData genericData)
        {
            TestSection.Arrange();

            var target = new UAOpcBinaryDts();
            var dataTypeDictionaryDescriptor = new UAModelNodeDescriptor();
            target.AddDataTypeDictionary(dataTypeDictionaryDescriptor, dictionaryByteArray);
            setupTargetAction?.Invoke(target);

            Exception validationException = genericData.TryValidate();
            AssertUtilities.WriteLine($"validationException: {validationException}");
            Trace.Assert(validationException == null);
            Trace.Assert(genericData.IsValid);

            TestSection.Act();

            ValueResult encodedValueResult = target.Encode(
                new UADataTypeDescriptor(dataTypeDictionaryDescriptor, dataTypeDescription),
                genericData);

            TestSection.Assert();

            AssertUtilities.WriteLine($"encodedValueResult: {encodedValueResult}");
            Assert.IsNotNull(encodedValueResult, "Encoded value result is a null reference.");
            AssertUtilities.WriteLine($"encodedValueResult.ErrorCode: {encodedValueResult.ErrorCode}");
            Exception exception = encodedValueResult.Exception;
            while (exception != null)
            {
                AssertUtilities.WriteLine($"exception.Message: {exception.Message}");
                exception = exception.InnerException;
            }
            AssertUtilities.WriteLine();
            return encodedValueResult;
        }

        [CanBeNull]
        static internal UAStandardBinaryIntrinsicDataTypeDictionary FindStandardBidtd([NotNull] UAOpcBinaryDts opcBinaryDts)
        {
            return (UAStandardBinaryIntrinsicDataTypeDictionary)opcBinaryDts.FindIntrinsicDataTypeDictionary(UAOpcBinaryDts.StandardTypesNamespace);
        }

        // for profiling
        static internal void RepeatedDecodeTest(
            [NotNull] byte[] dictionaryByteArray,
            [CanBeNull] object dataTypeDescription,
            [NotNull] byte[] extensionObjectBody,
            int repetitionCount)
        {
            var target = new UAOpcBinaryDts();
            var dataTypeDictionaryDescriptor = new UAModelNodeDescriptor();
            target.AddDataTypeDictionary(dataTypeDictionaryDescriptor, dictionaryByteArray);

            for (int i = 0; i < repetitionCount; i++)
            {
                ValueResult<GenericData> decodedValueResult = target.Decode(
                    new UADataTypeDescriptor(dataTypeDictionaryDescriptor, dataTypeDescription),
                    extensionObjectBody);
                Assert.IsTrue(decodedValueResult.Succeeded);
            }
        }

        // for profiling
        static internal void RepeatedEncodeTest(
            [NotNull] byte[] dictionaryByteArray,
            [CanBeNull] object dataTypeDescription,
            [NotNull] GenericData genericData,
            int repetitionCount)
        {
            var target = new UAOpcBinaryDts();
            var dataTypeDictionaryDescriptor = new UAModelNodeDescriptor();
            target.AddDataTypeDictionary(dataTypeDictionaryDescriptor, dictionaryByteArray);

            for (int i = 0; i < repetitionCount; i++)
            {
                ValueResult encodedValueResult = target.Encode(
                    new UADataTypeDescriptor(dataTypeDictionaryDescriptor, dataTypeDescription),
                    genericData);
                Assert.IsTrue(encodedValueResult.Succeeded);
            }
        }

        #endregion



        #region Static Private Methods

        #endregion
    }
}
